function [firstPeakHeight,secondPeakPos,secondPeakHeight,firstValleyPos,firstValleyHeight] = gofrstats(r,g,figureHandle,pValueInput)
% This function computes the heights and positions of the first two peaks, 
% and the first valley of g(r) 

% Default value of the smoothing parameter in the smoothing spline will be
% taken as 0.99999, which gives a fairly smooth curve. The p-value should
% ideally be larger 0.99 < p < 1, and can be explained here: 
% https://www.mathworks.com/help/curvefit/smoothing-splines.html
if nargin > 3
    pValue = pValueInput;
else
    pValue = 0.99999;
end

% Set up interpolation space and the smoothing spline
rSteps = linspace(min(r),max(r),1000);
splineFit = fit(r,g,'smoothingspline','smoothingParam',pValue,'normalize','on');

% Using findpeaks to find maxima and minima. Note that this is fundamentally
% limited by the graining of space, as we are not doing further
% interpolation
[maxVal,maxPosition] = findpeaks(feval(splineFit,rSteps),rSteps);
[minVal,minPosition] = findpeaks(-feval(splineFit,rSteps),rSteps);

% First peak is extrapolated as the value of the function at 1
firstPeakHeight = feval(splineFit,1);

% Other peaks and valleys taken from findpeaks
secondPeakPos = maxPosition(1);
secondPeakHeight = maxVal(1);
firstValleyPos = minPosition(1);
firstValleyHeight = -minVal(1);

figure(figureHandle)
clf; hold all
plot(r,g,'.k')
plot(rSteps,feval(splineFit,rSteps),'-r')
end