import numpy as np
from scipy.special import gamma
import monteCarloPCP as mc
import os.path

def powerRep(number,base,digits):
	# Turns a number into a representation in a base, cutoff such that the limit is base**digits
	# First, create an empty representation
	rep = [0]*digits
	for thisDigit in range(digits,0,-1):
		rep[thisDigit-1], number = divmod(number, base**(thisDigit-1))
	rep.reverse()
	return rep

def hyperBallCoefficient(dim):
	return (np.pi**(dim/2.0))/gamma(dim/2.0 + 1)

def generatedDCrystal(numCells,dim):
	if dim == 9:
		my_path = os.path.abspath(os.path.dirname(__file__))
		path = os.path.join(my_path, "d9N39366Pos.dat")
		print(path)
		crystalPos = mc.fileIO.load2DArray(path,dtype=float)
	else:
		nParticles = int((numCells**dim)/2)
		crystalPos = np.zeros((nParticles,dim),dtype=float)
		i = 0
		for j in range(2*nParticles): 
			testPosition = powerRep(j,numCells,dim)
			if np.sum(testPosition) % 2 == 0:
				crystalPos[i] = testPosition
				i += 1
		crystalPos /= float(numCells)
	sides = np.ones(dim)
	if dim == 9:
		sides *= 2
	return crystalPos, sides

def generatedDCrystalDBC(length,dim):
	sides = 2*np.ones(dim)
	nParticles = (2*length+1)**dim
	crystalPos = np.zeros((nParticles,dim),dtype=float)
	i = 0
	if length%2 == 0:
		maxSum = length
	else:
		maxSum = length+1
	for j in range(nParticles):
		testPosition = powerRep(j,2*length+1,dim) - length*np.ones(dim)
		testSum = np.sum(testPosition)
		if (testSum % 2 == 0 and testSum >= -length and testSum <= maxSum):
		# if (testSum % 2 == 0 and np.sum(np.round(testPosition/(length+1)))%2 == 0):
			crystalPos[i] = testPosition
			i += 1
	crystalPos = crystalPos[0:i,:]
	crystalPos /= float(length+1)
	nParticles = np.shape(crystalPos)[0]
	s = mc.MCPacking(dim,nParticles,maxPhi(dim))
	s.setGeometryType(mc.geometryEnum.Dn)
	s.setPositions(crystalPos)
	crystalPos = s.getPositions()
	todelete = s.uniqueParticles(1e-8)
	crystalPos = crystalPos[todelete==0,:]
	if length%2==1:
		added = np.zeros(dim)
		added[dim-1] = 1
		crystalPos = np.vstack((crystalPos,added))
	return crystalPos, sides

def maxPhi(dim):
	if (dim == 9):
		return 2 * hyperBallCoefficient(dim) * 2.0 **(-dim/2 -1)
	else:
		return hyperBallCoefficient(dim) * 2.0 **(-dim/2 -1)