#ifndef SD_H_
#define SD_H_

#include "MCPacking.h"
#include "PIMPLClasses.h"
#include "enumTypes.h"

#include <algorithm>
#include <tuple>
#include <omp.h>

//Implementation
MCPacking::SD::SD(MCPacking* ptr) : packing_(ptr) {

}

/********************** SD **************************/
//Set the mass for each degree of freedom
void MCPacking::SD::updatePosition(double stepSize, vector<double>& gradVector){
	// #pragma omp parallel for
	packing_->translate(stepSize,gradVector);
	// packing_->calcNeighborsCut(1.0);
}

//Functions related to the minimizeFIRE
void MCPacking::SD::initMinimizer(double initialStepSize_, double maxForceRescale_, double f_stepSizeDecrease_, bool checkGradient_, double gradientCheck_, int recheckRate_, int sdLastDecrease_, int sdLastIncrease_, double time_) {
	initialStepSize = initialStepSize_;
	packing_->pi_->maxForceRescale = maxForceRescale_;
	checkGradient = checkGradient_;
	gradientCheck = gradientCheck_;
	f_stepSizeDecrease = f_stepSizeDecrease_;
	sdLastDecrease = sdLastDecrease_;
	sdLastIncrease = sdLastIncrease_;
	recheckRate = recheckRate_;
	time = time_;
	//Create energy and force arrays
	packing_->initForceEnergy();
	packing_->calcForceEnergy();
	if (packing_->pi_->simControl.potentialType == simControlStruct::potentialEnum::log || packing_->pi_->simControl.potentialType == simControlStruct::potentialEnum::wca || packing_->pi_->simControl.potentialType == simControlStruct::potentialEnum::inversePower) {
		packing_->pi_->rescaleForces(packing_->pi_->maxForceRescale);
	}
}

void MCPacking::SD::minimizerLoop(){
	vector<double> lastUnwrappedPositions = packing_->pi_->unwrappedPos;
	// vector<double> lastPositions = packing_->pi_->pos;
	vector<double> lastTotalDisplacement = packing_->pi_->totalDisplacement;
	vector<double> lastDistTravelledSq = packing_->pi_->distTravelledSq;
	bool goodStep = false;
	double stepSize = initialStepSize;
	// packing_->calcForceEnergy();
	vector<double> lastForces = packing_->pi_->netForce;
	// stepSize = initialStepSize*packing_->getMeanRadius()/packing_->getMaxUnbalancedForce();
	// vector<long> indices1; 
	// vector<long> indices2; 
	// vector<double> overlapsData;
	// tie(indices1,indices2,overlapsData) = packing_->getSmallOverlaps(0.0);
	// double maxOverlap = *max_element(overlapsData.begin(),overlapsData.end());
	// stepSize = initialStepSize*maxOverlap*2.0*packing_->getMeanRadius()/packing_->getMaxUnbalancedForce();
	stepSize = initialStepSize * 2.0*packing_->getMeanRadius()/packing_->getMaxUnbalancedForce();
	// stepSize = initialStepSize*2.0*packing_->getMeanRadius()*packing_->getMaxUnbalancedForce();
	double lastEnergy = packing_->getEnergy();
	int numRecalcs = 0;
	double gradientNorm;
	while (!goodStep){
		updatePosition(stepSize,lastForces); //Move the system forward, based on the previous forces
		packing_->calcForceEnergy();
		// cout << "step " << stepSize << endl;
		// cout << "energy " << packing_->getEnergy() << endl;
		gradientNorm = inner_product(lastForces.begin(), lastForces.end(), packing_->pi_->netForce.begin(), 0.0)/sqrt(inner_product(packing_->pi_->netForce.begin(), packing_->pi_->netForce.end(), packing_->pi_->netForce.begin(), 0.0)*inner_product(lastForces.begin(), lastForces.end(), lastForces.begin(), 0.0));
		// cout << "unbalanced force " << packing_->getMaxUnbalancedForce() << endl;
		// if ((packing_->getEnergy() - lastEnergy > packing_->pi_->numParticles*1e-13) || (checkGradient && inner_product(lastForces.begin(), lastForces.end(), packing_->pi_->netForce.begin(), 0.0) < gradientCheck)){
		// if ((checkGradient && gradientNorm < gradientCheck) && stepSize/(2.0*packing_->getMeanRadius()/packing_->getMaxUnbalancedForce()) > 1e-15 && packing_->getMaxUnbalancedForce() / packing_->getMeanRadius() > 1e-13){
		if ((checkGradient && gradientNorm < gradientCheck) && stepSize/(2.0*packing_->getMeanRadius()/packing_->getMaxUnbalancedForce()) > 1e-15 && packing_->getMaxUnbalancedForce() / packing_->getMeanRadius() > 1e-13){
		// if ((checkGradient && gradientNorm < gradientCheck) && stepSize > 1e-15/packing_->getMeanRadius()){
		// if ((checkGradient && gradientNorm < gradientCheck)){
			stepSize *= f_stepSizeDecrease;
			packing_->setPositions(lastUnwrappedPositions);
			packing_->pi_->totalDisplacement = lastTotalDisplacement;
			packing_->pi_->distTravelledSq = lastDistTravelledSq;
			numRecalcs++;
		}else{
			goodStep = true;
		}
	}
	if (numRecalcs == 0){
		sdLastIncrease++;
		sdLastDecrease = 0;
		if (sdLastIncrease == recheckRate){
			initialStepSize /= f_stepSizeDecrease;
			sdLastIncrease = 0;
			// cout << "increasing " << initialStepSize << " energy " << packing_->getEnergy() << endl;
		}
	}else{
		sdLastIncrease = 0;
		sdLastDecrease++;
		if (sdLastDecrease == recheckRate){
			initialStepSize *= f_stepSizeDecrease;
			sdLastDecrease = 0;
			// cout << "decreasing " << initialStepSize << " energy " << packing_->getEnergy() << endl;
		}
	}
	// cout << sdLastDecrease << " " << sdLastIncrease << endl;
	if (packing_->pi_->simControl.potentialType == simControlStruct::potentialEnum::log || packing_->pi_->simControl.potentialType == simControlStruct::potentialEnum::wca || packing_->pi_->simControl.potentialType == simControlStruct::potentialEnum::inversePower) {
		packing_->pi_->rescaleForces(packing_->pi_->maxForceRescale);
	}
	time += stepSize;
}


#endif /* SD_H_ */