#include "enumTypes.h"
#include "MCPacking.h"
#include <vector>
#include <tuple>
#include <functional>
#include <utility>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/numpy.h>
#include <pybind11/functional.h>
#include <pybind11/chrono.h>
#include <pybind11/complex.h>
#include <omp.h>

namespace py = pybind11;

PYBIND11_MODULE(libmonteCarloPCP, m) {
    pybind11::enum_<simControlStruct::neighborEnum>(m, "neighborEnum")
    .value("allToAll", simControlStruct::neighborEnum::allToAll)
    .value("nList", simControlStruct::neighborEnum::nList)
    .value("hex2D", simControlStruct::neighborEnum::hex2D);

    pybind11::enum_<simControlStruct::distanceEnum>(m, "distanceEnum")
    .value("MK", simControlStruct::distanceEnum::MK)
    .value("nonAdditive", simControlStruct::distanceEnum::nonAdditive)
    .value("normal", simControlStruct::distanceEnum::normal);

    pybind11::enum_<simControlStruct::geometryEnum>(m, "geometryEnum")
    .value("normal", simControlStruct::geometryEnum::normal)
    .value("nonSquare", simControlStruct::geometryEnum::nonSquare)
    .value("leesEdwards", simControlStruct::geometryEnum::leesEdwards)
    .value("E8", simControlStruct::geometryEnum::E8)
    .value("Dn", simControlStruct::geometryEnum::Dn);

    pybind11::enum_<simControlStruct::potentialEnum>(m, "potentialEnum")
    .value("contactPower", simControlStruct::potentialEnum::contactPower)
    .value("log", simControlStruct::potentialEnum::log)
    .value("wca", simControlStruct::potentialEnum::wca)
    .value("inversePower", simControlStruct::potentialEnum::inversePower);

    pybind11::enum_<simControlStruct::constraintEnum>(m, "constraintEnum")
    .value("normal", simControlStruct::constraintEnum::normal)
    .value("strainControl", simControlStruct::constraintEnum::strainControl)
    .value("stressControl", simControlStruct::constraintEnum::stressControl);

    pybind11::enum_<simControlStruct::einsteinEnum>(m, "einsteinEnum")
    .value("none", simControlStruct::einsteinEnum::none)
    .value("spring", simControlStruct::einsteinEnum::spring)
    .value("springUntethered", simControlStruct::einsteinEnum::springUntethered)
    .value("periodic", simControlStruct::einsteinEnum::periodic)
    .value("periodicUmbrella", simControlStruct::einsteinEnum::periodicUmbrella)
    .value("periodicUmbrellaExp", simControlStruct::einsteinEnum::periodicUmbrellaExp);

    pybind11::enum_<simControlStruct::particleEnum>(m, "particleEnum")
    .value("sphere", simControlStruct::particleEnum::sphere)
    .value("cube", simControlStruct::particleEnum::cube);

    pybind11::enum_<simControlStruct::ensembleEnum>(m, "ensembleEnum")    
    .value("nvt", simControlStruct::ensembleEnum::nvt)
    .value("npt", simControlStruct::ensembleEnum::npt);

    pybind11::enum_<simControlStruct::monteCarloEnum>(m, "monteCarloEnum")
    .value("metropolis", simControlStruct::monteCarloEnum::metropolis)
    .value("ecmc", simControlStruct::monteCarloEnum::ecmc);


    py::class_<MCPacking>(m, "MCPacking")
        .def(py::init<int,long,double>() )

        .def("anyOverlap",&MCPacking::anyOverlap)
        .def("calcContacts", &MCPacking::calcContacts)
        .def("calcEnergy", &MCPacking::calcEnergy)
        .def("calcForceEnergy", &MCPacking::calcForceEnergy)
        .def("calcGofR", &MCPacking::calcGofR)
        .def("calcGRBinVolume", &MCPacking::calcGRBinVolume)
        .def("calcMCOverlap", &MCPacking::calcMCOverlap)
        .def("calcNeighborsCut", &MCPacking::calcNeighborsCut)
        .def("calcOverlap", &MCPacking::calcOverlap)
        .def("createD9Plus", &MCPacking::createD9Plus)
        .def("fastRound", &MCPacking::fastRound)
        // .def("fastRound", static_cast<double (MCPacking::*)(double)>(&MCPacking::fastRound))
        // .def("fastRound", static_cast<float (MCPacking::*)(float)>(&MCPacking::fastRound))
        // .def("fastRound", static_cast<double (MCPacking::*)(double)>(&MCPacking::fastRound))
        // .def("fastRound", static_cast<float (MCPacking::*)(float)>(&MCPacking::fastRound))
        .def("getAcceptanceRatio", &MCPacking::getAcceptanceRatio)
        .def("getBoxSize", &MCPacking::getBoxSize)
        .def("getCenterOfMassShift", &MCPacking::getCenterOfMassShift)
        .def("getConstraint", &MCPacking::getConstraint)
        .def("getConstraintMagnitude", &MCPacking::getConstraintMagnitude)
        .def("getConstraintType", &MCPacking::getConstraintType)
        .def("getContacts", &MCPacking::getContacts)
        .def("getContactVectors", &MCPacking::getContactVectors)
        .def("getCrystalPositions", &MCPacking::getCrystalPositions)
        .def("getDistanceType", &MCPacking::getDistanceType)
        .def("getD9Shift", &MCPacking::getD9Shift)
        .def("getECMCLength", &MCPacking::getECMCLength)
        .def("getECMCPressure", &MCPacking::getECMCPressure)
        .def("getEinsteinEnergy", &MCPacking::getEinsteinEnergy)
        .def("getEinsteinLambda", &MCPacking::getEinsteinLambda)
        .def("getEinsteinType", &MCPacking::getEinsteinType)
        .def("getEnsembleType", &MCPacking::getEnsembleType)
        .def("getEinsteinWeightedConstant", &MCPacking::getEinsteinWeightedConstant)
        .def("getEnergy", &MCPacking::getEnergy)
        .def("getEnergyVector", &MCPacking::getEnergyVector)
        .def("getExternalForce", &MCPacking::getExternalForce)
        .def("getFIREDamping", &MCPacking::getFIREDamping)
        .def("getFIREIteration", &MCPacking::getFIREIteration)
        .def("getFIRETimeStep", &MCPacking::getFIRETimeStep)
        .def("getGeometryType", &MCPacking::getGeometryType)
        // .def("getHessian", &MCPacking::getHessian)
        .def("getInversePower", &MCPacking::getInversePower)
        .def("getInversePowerCutoff", &MCPacking::getInversePowerCutoff)
        .def("_getLastPositions", &MCPacking::_getLastPositions)
        .def("getLEshift", &MCPacking::getLEshift)
        .def("getLogZero", &MCPacking::getLogZero)
        .def("getMaxDisplacement", &MCPacking::getMaxDisplacement)
        .def("getMaxUnbalancedForce", &MCPacking::getMaxUnbalancedForce)
        .def("getMaxVolumeShift", &MCPacking::getMaxVolumeShift)
        .def("getMCMaxDisplacement", &MCPacking::getMCMaxDisplacement)
        .def("getMCMaxDisplacementCM", &MCPacking::getMCMaxDisplacementCM)
        .def("getMCMaxDisplacementD9Relative", &MCPacking::getMCMaxDisplacementD9Relative)
        .def("getMCMCBlockedDirection", &MCPacking::getMCMCBlockedDirection)
        .def("getMCOverlap", &MCPacking::getMCOverlap)
        .def("getMeanRadius", &MCPacking::getMeanRadius)
        .def("getMinGap", &MCPacking::getMinGap)
        // .def("getMKLambdas", &MCPacking::getMKLambdas)
        .def("getMKLambdaValue", &MCPacking::getMKLambdaValue)
        .def("getMKStartPoint", &MCPacking::getMKStartPoint)
        .def("getMonteCarloType", &MCPacking::getMonteCarloType)
        .def("getMTState", &MCPacking::getMTState)
        .def("getNDim", &MCPacking::getNDim)
        .def("getNeighbors", &MCPacking::getNeighbors)
        .def("getNeighborType", &MCPacking::getNeighborType)
        .def("getNetForces", &MCPacking::getNetForces)
        .def("getNonAdditivity", &MCPacking::getNonAdditivity)
        .def("getNumParticles", &MCPacking::getNumParticles)
        .def("getNumRows", &MCPacking::getNumRows)
        .def("getParticleType", &MCPacking::getParticleType)
        .def("getPerParticleStressTensor",&MCPacking::getPerParticleStressTensor)
        .def("getPhi", &MCPacking::getPhi)
        .def("getPositions", &MCPacking::getPositions)
        .def("getPotentialPower", &MCPacking::getPotentialPower)
        .def("getPotentialType", &MCPacking::getPotentialType)
        .def("getRadii", &MCPacking::getRadii)
        .def("getRadSum", &MCPacking::getRadSum)
        .def("getRandomGaussianVector", &MCPacking::getRandomGaussianVector)
        .def("getRandomVector", &MCPacking::getRandomVector)
        .def("getSDLastDecrease", &MCPacking::getSDLastDecrease)
        .def("getSDLastIncrease", &MCPacking::getSDLastIncrease)
        .def("getSDStepSize", &MCPacking::getSDStepSize)
        .def("getSDTime", &MCPacking::getSDTime)
        .def("getSmallGaps", &MCPacking::getSmallGaps)
        .def("getSmallOverlaps", &MCPacking::getSmallOverlaps)
        .def("getSpringConstant", &MCPacking::getSpringConstant)
        .def("getSumOfMaxDisplacements", &MCPacking::getSumOfMaxDisplacements)
        .def("getSumOfMaxMKDisplacements", &MCPacking::getSumOfMaxMKDisplacements)
        .def("getTargetPressure", &MCPacking::getTargetPressure)
        .def("getTauAlpha", &MCPacking::getTauAlpha)
        .def("getTemperature", &MCPacking::getTemperature)
        .def("getTotalDisplacement", &MCPacking::getTotalDisplacement)
        .def("getUnwrappedPositions", &MCPacking::getUnwrappedPositions)
        .def("getVolume", &MCPacking::getVolume)
        .def("initFIREMinimizer", &MCPacking::initFIREMinimizer)
        .def("initSDMinimizer", &MCPacking::initSDMinimizer)
        .def("MCStep", &MCPacking::MCStep)
        .def("MCStepAspectRatio", &MCPacking::MCStepAspectRatio)
        .def("MCStepCM", &MCPacking::MCStepCM)
        .def("MCStepD9Relative", &MCPacking::MCStepD9Relative)
        .def("MCStepVolume", &MCPacking::MCStepVolume)
        .def("MCStepVolumeStretch", &MCPacking::MCStepVolumeStretch)
        .def("minimizerLoopFIRE", &MCPacking::minimizerLoopFIRE)
        .def("minimizerLoopSD", &MCPacking::minimizerLoopSD)
        .def("pbcDisplacement", &MCPacking::pbcDisplacementOutside)
        .def("pbcDistance", &MCPacking::pbcDistanceOutside)
        .def("pbcParticleDisplacement", &MCPacking::pbcParticleDisplacement)
        .def("pbcParticleDistance", &MCPacking::pbcParticleDistance)
        .def("removeParticles", &MCPacking::removeParticles)
        .def("rescaleForces", &MCPacking::rescaleForces)
        .def("resetCenterOfMass", &MCPacking::resetCenterOfMass)
        .def("resetDisplacement", &MCPacking::resetDisplacement)
        .def("resetFIREMaxTimeStep", &MCPacking::resetFIREMaxTimeStep)
        // .def("save", &MCPacking::save)
        .def("setBoxSize", &MCPacking::setBoxSize)
        .def("setConstraint", &MCPacking::setConstraint)
        .def("setConstraintType", &MCPacking::setConstraintType)
        .def("setCrystalPositions", &MCPacking::setCrystalPositions)
        .def("setDistanceType", &MCPacking::setDistanceType)
        .def("setD9Shift", &MCPacking::setD9Shift)
        .def("setECMCLength", &MCPacking::setECMCLength)
        .def("setEinsteinLambda", &MCPacking::setEinsteinLambda)
        .def("setEinsteinType", &MCPacking::setEinsteinType)
        .def("setEnsembleType", &MCPacking::setEnsembleType)
        .def("setExternalForce", &MCPacking::setExternalForce)
        .def("setFIREDamping", &MCPacking::setFIREDamping)
        .def("setFIREIteration", &MCPacking::setFIREIteration)
        .def("setFIRETimeStep", &MCPacking::setFIRETimeStep)
        .def("setGeometryType", &MCPacking::setGeometryType)
        .def("setInitialPositions", &MCPacking::setInitialPositions)
        .def("setInversePower", &MCPacking::setInversePower)
        .def("setInversePowerConstants", &MCPacking::setInversePowerConstants)
        .def("_setLastPositions", &MCPacking::_setLastPositions)
        .def("setLatticeSpacing", &MCPacking::setLatticeSpacing)
        .def("setLEshift", &MCPacking::setLEshift)
        .def("setLogZero", &MCPacking::setLogZero)
        .def("setMaxForceRescale", &MCPacking::setMaxForceRescale)
        .def("setMaxVolumeShift", &MCPacking::setMaxVolumeShift)
        .def("setMCMaxDisplacement", &MCPacking::setMCMaxDisplacement)
        .def("setMCMaxDisplacementCM", &MCPacking::setMCMaxDisplacementCM)
        .def("setMCMaxDisplacementD9Relative", &MCPacking::setMCMaxDisplacementD9Relative)
        .def("setMCMCBlockedDirection", &MCPacking::setMCMCBlockedDirection)
        .def("setMCOverlap", &MCPacking::setMCOverlap)
        // .def("setMKLambdas", &MCPacking::setMKLambdas)
        .def("setMKLambdaValue", &MCPacking::setMKLambdaValue)
        .def("setMKStartPoint", &MCPacking::setMKStartPoint)
        .def("setMonteCarloType", &MCPacking::setMonteCarloType)
        .def("setMTState", &MCPacking::setMTState)
        .def("setNDim", &MCPacking::setNDim)
        .def("setNeighborType", &MCPacking::setNeighborType)
        .def("setNonAdditivity", &MCPacking::setNonAdditivity)
        .def("setNumParticles", &MCPacking::setNumParticles)
        .def("setNumRows", &MCPacking::setNumRows)
        .def("setParticleType", &MCPacking::setParticleType)
        .def("setPhi", &MCPacking::setPhi)
        .def("setPositions", &MCPacking::setPositions)
        .def("setPotentialPower", &MCPacking::setPotentialPower)
        .def("setPotentialType", &MCPacking::setPotentialType)
        .def("setRadii", &MCPacking::setRadii)
        .def("setRandomMKLambdas", &MCPacking::setRandomMKLambdas)
        .def("setRandomPositions", &MCPacking::setRandomPositions)
        .def("setRandomSeed", &MCPacking::setRandomSeed)
        .def("setSDLastDecrease", &MCPacking::setSDLastDecrease)
        .def("setSDLastIncrease", &MCPacking::setSDLastIncrease)
        .def("setSDStepSize", &MCPacking::setSDStepSize)
        .def("setTargetPressure", &MCPacking::setTargetPressure)
        .def("setTauAlpha", &MCPacking::setTauAlpha)
        .def("setTemperature", &MCPacking::setTemperature)
        // .def("softMKThermalizationStep", &MCPacking::softMKThermalizationStep)
        .def("translate", &MCPacking::translate)
        .def("uniqueParticles", &MCPacking::uniqueParticles)
        .def("zeroVelocities", &MCPacking::zeroVelocities);
}
