import monteCarloPCP as mc
import numpy as np
import sys
import os
import glob
import time
import Resources.crystalStructures.e8Crystal as e8
import Resources.crystalStructures.e7Crystal as e7
import Resources.crystalStructures.e6Crystal as e6
import Resources.crystalStructures.dDCrystal as dD
import Resources.crystalStructures.p10cCrystal as p10c


# addMSD averages two MSD runs with different lengths
def addMSD(thisMsd,thisAlpha2,thisMsdDivide,msd,alpha2,msdDivide):
	if len(thisMsd) < len(msd):
		newMsd = msd.copy()
		newMsd[:len(thisMsd)] += thisMsd
		newAlpha2 = alpha2.copy()
		newAlpha2[:len(thisAlpha2)] += thisAlpha2
		newMsdDivide = msdDivide.copy()
		newMsdDivide[:len(thisMsdDivide)] += thisMsdDivide
	else:
		newMsd = thisMsd.copy()
		newMsd[:len(msd)] += msd
		newAlpha2 = thisAlpha2.copy()
		newAlpha2[:len(alpha2)] += alpha2
		newMsdDivide = thisMsdDivide
		newMsdDivide[:len(msdDivide)] += msdDivide
	return newMsd, newAlpha2, newMsdDivide

# Input parameters
dim = int(sys.argv[1])
nParticles = int(sys.argv[2])
rhoString = sys.argv[3]
numR = int(sys.argv[4])
saveFreq = int(sys.argv[5])
numAvgs = int(sys.argv[6])
mcOverlapString = sys.argv[7]
numTau = int(sys.argv[8])
basedir = sys.argv[9]

useFixedDisp = 1
fullDist = 1

# The following block simply sets up the simulations with D boundary conditions 
# and correct neighbor list
if dim==6:
	maxPhi = e6.maxPhi()
elif dim==8:
	maxPhi = dD.maxPhi(dim)
elif dim==9:
	maxPhi = dD.maxPhi(dim)
elif dim == 7:
	maxPhi = e7.maxPhi()
else:
	maxPhi = dD.maxPhi(dim)

rho = np.float64(rhoString)
volPoints = 1000000
saveDir = basedir + '/' + mcOverlapString + '/' + str(dim) + '/' + rhoString + '/' + str(nParticles) + '/'
tempDir = saveDir + '/temp/'

p = mc.MCPacking(dim,nParticles,0.1)

if dim == 3:
	p.setBiRadii()

boxVolume = 1

boxVolume = 2
p.setGeometryType(mc.geometryEnum.Dn)

phi = rho * p.hyperBallCoefficient(dim)/(2.0**dim)
p.setPhi(phi)
maxRad = (maxPhi * boxVolume/(nParticles * p.hyperBallCoefficient(dim)))**(1/dim)
maxDist = ((maxRad/p.getRadii()[0] - 1) * 4 + 1) * 2 * p.getRadii()[0]
cutDist = 1

# Start with random positions, energy mnimized
p.setRandomPositions()
p.calcNeighborsCut(cutDist)

p.calcForceEnergy()
p.minimizeFIRE(criticalForce = 1e-12)
p.calcForceEnergy()

p.setMCOverlap(np.float64(mcOverlapString))

maxMCDispFixed = np.array([0,0,0.110,0.060,0.040,0.033,0.027,0.024,0.021,0.019,0.0170,0.016,0.0142,0.013,0.0121,0.011,0.0106,0.008])

print("volume is " + str(p.getVolume()))

minRadius = np.min(p.getRadii())
# Calculate what max displacement should be

os.makedirs(saveDir,exist_ok=True)


# If an initial equilibrated data set has been used, use that, otherwise make sure
# that the energy minimized data is properly thermalized.
if os.path.exists(tempDir):
	with open(saveDir + '/neighborCutDistance.dat', 'r') as file:
		neighborCutDistance = np.float(file.read())
	with open(saveDir + '/maxMCDisp.dat', 'r') as file:
		maxMCDisp = np.float(file.read())
	p.load(tempDir)
	msd = mc.fileIO.loadArray(saveDir + '/msd.dat',dtype=np.float64)
	alpha2 = mc.fileIO.loadArray(saveDir + '/alpha2.dat',dtype=np.float64)
	msdDivide = mc.fileIO.loadArray(saveDir + 'msdDivide.dat',dtype=np.float64)
else:
	initPos = p.getUnwrappedPositions()
	msd = np.array([],dtype=np.float64)
	alpha2 = np.array([],dtype=np.float64)
	msdDivide = np.array([],dtype=np.float64)
	if useFixedDisp == 1:
		numSteps, tauAlpha, neighborCutDistance, maxMCDisp, msd, alpha2, _ = p.thermalizePacking(verbose=True,numTrials=100,numCosts=10,numTau=2,calcMSD=True)
		p.resetCenterOfMass()
		numSteps, tauAlpha, neighborCutDistance, maxMCDisp, msd, alpha2, _ = p.thermalizePacking(verbose=True,maxMCDisp=maxMCDispFixed[dim],firstPass=False,numTrials=100,numCosts=30,numTau=numTau,calcMSD=True)
	else:
		p.resetCenterOfMass()
		numSteps, tauAlpha, neighborCutDistance, maxMCDisp, msd, alpha2, _ = p.thermalizePacking(verbose=True,numTrials=100,numCosts=5,numTau=numTau,calcMSD=True)
	msdDivide = 0 * msd + 1

	if not os.path.exists(saveDir + 'msdDivide.dat'):
		mc.fileIO.saveArray(saveDir + 'msd.dat',msd)
		mc.fileIO.saveArray(saveDir + 'alpha2.dat',alpha2)
		mc.fileIO.saveArray(saveDir + 'msdDivide.dat',msdDivide)
	p.save(tempDir)
	with open(saveDir + '/maxMCDisp.dat', 'w') as file:
		file.write(str(maxMCDisp) + '\n')
	with open(saveDir + '/neighborCutDistance.dat', 'w') as file:
		file.write(str(neighborCutDistance) + '\n')
	with open(saveDir + '/tauAlpha.dat', 'w') as file:
		file.write(str(int(tauAlpha)) + '\n')

# Set up the apparatus to save gofr and msd
rad = p.getRadii()[0]
g = np.zeros(numR)
r = np.zeros(numR)
p.setRandomSeed(np.random.randint(2**16))
start = len(glob.glob(saveDir + '/[0-9]*/'))
p.calcNeighborsCut(neighborCutDistance);
p.calcForceEnergy()
p.resetDisplacement()
saveNum = start
tauList = np.zeros(saveFreq)
DList = np.zeros(saveFreq)
i = int(saveFreq*saveNum)+1

msd = np.array([],dtype=np.float64)
alpha2 = np.array([],dtype=np.float64)
msdDivide = np.array([],dtype=np.float64)

# Run one equilibration and save it
while (i < numAvgs+1):
	p.calcNeighborsCut(neighborCutDistance)
	p.resetCenterOfMass()
	initPos = p.getUnwrappedPositions()
	_,tauAlpha,_,_,thisMsd,thisAlpha2,_ = p.thermalizePacking(numTau=numTau,maxMCDisp=maxMCDisp,neighborCutDistance=neighborCutDistance,firstPass=False,calcMSD=True)
	thisMsdDivide = np.ones(len(thisMsd))
	msd,alpha2,msdDivide = addMSD(thisMsd,thisAlpha2,thisMsdDivide,msd,alpha2,msdDivide)
	finalPos = p.getUnwrappedPositions()
	delta = finalPos - initPos - p.getCenterOfMassShift()
	msd1 = np.mean(np.sum(delta**2,1))/(rad**2)
	diffConst = msd1/(2*dim*3*tauAlpha)
	p.calcNeighborsCut(10)
	if fullDist == 1:
		thisg,thisr = p.gofr(numR,minRad = p.getRadii()[0]*2,numVolumePoints = volPoints,logBin = False)
	else:
		thisg,thisr = p.gofr(numR,maxRad = maxDist,minRad = p.getRadii()[0]*2,numVolumePoints = volPoints,logBin = False)
	thisg[np.isnan(thisg)] = 0
	thisg[np.isinf(thisg)] = 0
	r = thisr/(2*p.getRadii()[0])
	g = g+thisg
	tauList[i%saveFreq] = tauAlpha
	DList[i%saveFreq] = diffConst
	if (i%saveFreq == 0):
		saveNum = len(glob.glob(saveDir + '/[0-9]*/'))
		os.makedirs(saveDir + '/' + str(saveNum),exist_ok=True)
		mc.fileIO.save2DArray(saveDir + '/' + str(saveNum) + '/gofr.dat',np.stack((thisr/(2*p.getRadii()[0]),g/saveFreq),axis=1))
		g = np.zeros(numR)
		saveNum += 1
		for ii in range(len(tauList)):
			with open(saveDir + '/tauAlpha.dat', 'a') as file:
				file.write(str(int(tauList[ii])) + '\n')
		for ii in range(len(DList)):
			with open(saveDir + '/diffConst.dat', 'a') as file:
				file.write(str(DList[ii]) + '\n')
		oldMsd = mc.fileIO.loadArray(saveDir + '/msd.dat',dtype=np.float64)
		oldAlpha2 = mc.fileIO.loadArray(saveDir + '/alpha2.dat',dtype=np.float64)
		oldMsdDivide = mc.fileIO.loadArray(saveDir + 'msdDivide.dat',dtype=np.float64)
		msd,alpha2,msdDivide = addMSD(msd,alpha2,msdDivide,oldMsd,oldAlpha2,oldMsdDivide)
		mc.fileIO.saveArray(saveDir + 'msd.dat',msd)
		mc.fileIO.saveArray(saveDir + 'alpha2.dat',alpha2)
		mc.fileIO.saveArray(saveDir + 'msdDivide.dat',msdDivide)
		tauList = np.zeros(saveFreq)
		DList = np.zeros(saveFreq)
		p.save(tempDir)
		msd = np.array([],dtype=np.float64)
		alpha2 = np.array([],dtype=np.float64)
		msdDivide = np.array([],dtype=np.float64)
	i+=1


